"""Reference implementation of the CWL standards."""

__author__ = "pamstutz@veritasgenetics.com"

CWL_CONTENT_TYPES = [
    "text/plain",
    "application/json",
    "text/vnd.yaml",
    "text/yaml",
    "text/x-yaml",
    "application/x-yaml",
    "application/octet-stream",
]
