from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class OWL(DefinedNamespace):
    allValuesFrom: URIRef
    annotatedProperty: URIRef
    annotatedSource: URIRef
    annotatedTarget: URIRef
    assertionProperty: URIRef
    cardinality: URIRef
    complementOf: URIRef
    datatypeComplementOf: URIRef
    differentFrom: URIRef
    disjointUnionOf: URIRef
    disjointWith: URIRef
    distinctMembers: URIRef
    equivalentClass: URIRef
    equivalentProperty: URIRef
    hasKey: URIRef
    hasSelf: URIRef
    hasValue: URIRef
    intersectionOf: URIRef
    inverseOf: URIRef
    maxCardinality: URIRef
    maxQualifiedCardinality: URIRef
    members: URIRef
    minCardinality: URIRef
    minQualifiedCardinality: URIRef
    onClass: URIRef
    onDataRange: URIRef
    onDatatype: URIRef
    onProperties: URIRef
    onProperty: URIRef
    oneOf: URIRef
    propertyChainAxiom: URIRef
    propertyDisjointWith: URIRef
    qualifiedCardinality: URIRef
    sameAs: URIRef
    someValuesFrom: URIRef
    sourceIndividual: URIRef
    targetIndividual: URIRef
    targetValue: URIRef
    unionOf: URIRef
    withRestrictions: URIRef
    AllDifferent: URIRef
    AllDisjointClasses: URIRef
    AllDisjointProperties: URIRef
    Annotation: URIRef
    AnnotationProperty: URIRef
    AsymmetricProperty: URIRef
    Axiom: URIRef
    Class: URIRef
    DataRange: URIRef
    DatatypeProperty: URIRef
    DeprecatedClass: URIRef
    DeprecatedProperty: URIRef
    FunctionalProperty: URIRef
    InverseFunctionalProperty: URIRef
    IrreflexiveProperty: URIRef
    NamedIndividual: URIRef
    NegativePropertyAssertion: URIRef
    ObjectProperty: URIRef
    Ontology: URIRef
    OntologyProperty: URIRef
    ReflexiveProperty: URIRef
    Restriction: URIRef
    SymmetricProperty: URIRef
    TransitiveProperty: URIRef
    backwardCompatibleWith: URIRef
    deprecated: URIRef
    incompatibleWith: URIRef
    priorVersion: URIRef
    versionInfo: URIRef
    Nothing: URIRef
    Thing: URIRef
    bottomDataProperty: URIRef
    topDataProperty: URIRef
    bottomObjectProperty: URIRef
    topObjectProperty: URIRef
    imports: URIRef
    versionIRI: URIRef
