from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class FOAF(DefinedNamespace):
    account: URIRef
    accountName: URIRef
    accountServiceHomepage: URIRef
    age: URIRef
    based_near: URIRef
    birthday: URIRef
    currentProject: URIRef
    depiction: URIRef
    depicts: URIRef
    dnaChecksum: URIRef
    familyName: URIRef
    family_name: URIRef
    firstName: URIRef
    focus: URIRef
    fundedBy: URIRef
    geekcode: URIRef
    gender: URIRef
    givenName: URIRef
    givenname: URIRef
    holdsAccount: URIRef
    img: URIRef
    interest: URIRef
    knows: URIRef
    lastName: URIRef
    made: URIRef
    maker: URIRef
    member: URIRef
    membershipClass: URIRef
    myersBriggs: URIRef
    name: URIRef
    nick: URIRef
    page: URIRef
    pastProject: URIRef
    phone: URIRef
    plan: URIRef
    primaryTopic: URIRef
    publications: URIRef
    schoolHomepage: URIRef
    sha1: URIRef
    skypeID: URIRef
    status: URIRef
    surname: URIRef
    theme: URIRef
    thumbnail: URIRef
    tipjar: URIRef
    title: URIRef
    topic: URIRef
    topic_interest: URIRef
    workInfoHomepage: URIRef
    workplaceHomepage: URIRef
    Agent: URIRef
    Document: URIRef
    Group: URIRef
    Image: URIRef
    LabelProperty: URIRef
    OnlineAccount: URIRef
    OnlineChatAccount: URIRef
    OnlineEcommerceAccount: URIRef
    OnlineGamingAccount: URIRef
    Organization: URIRef
    Person: URIRef
    PersonalProfileDocument: URIRef
    Project: URIRef
    aimChatID: URIRef
    homepage: URIRef
    icqChatID: URIRef
    isPrimaryTopicOf: URIRef
    jabberID: URIRef
    logo: URIRef
    mbox: URIRef
    mbox_sha1sum: URIRef
    msnChatID: URIRef
    openid: URIRef
    weblog: URIRef
    yahooChatID: URIRef
